#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <sys/select.h>
#include <pthread.h>
#include <errno.h>
#include <pcap.h>
#include <fcntl.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <net/if.h>
#include <time.h>
#include <locale.h>
#include <netinet/if_ether.h>
#include <signal.h>

#define NUM_THREAD 3
#define MAX_THREAD 12
#define MAX_STRING 255
#define MAX_PORT 1024
#define BUF_LEN 500
#define DONE 1
#define SYN_SEC 5000
#define SYN_WAIT 30
#define THREAD_WAIT usleep(300)
#define ICMP_WAIT 1
#define PORT_WAIT 500000

struct options {
       char logging;
       char singlehost;
       char fromlist; 
       char pingnet;
       char subscan;
       char verbose;
       char interface;
       char synscan;
       char passivemode;
       char fingermode;
       char bpf;
       char headers;
       char payload;
       char *local_net;
       char *filter;
       char *logfile;
       char *hostlist;
       char *host;
       char os[MAX_STRING];
       unsigned long ip;
       unsigned int class;
       unsigned long myip;
       int threads;
       int banner;      
       int portspeed; 
       FILE *logfd; 
   };

struct host {
	char *name;
	unsigned long ip;
        int ports[1024];
        char *os;
        char **vuln;
                                              /* Each host always host its */
        struct bannerz {                      /* own thread and socket */
               char banner[MAX_STRING];
               int port;
            } banners;
        
        struct sockets {
               int sockfd;    // dont think will use this
               int flag;      // but we do use flag
         } socket;
     
    pthread_t tid;
};

struct portz {
       int max_port;
       int port[MAX_PORT];
       int open_port[MAX_PORT];
} p_list;

extern struct bpf_program fp;
extern char errbuf[BUF_LEN];
struct local {
        bpf_u_int32 mask;
        bpf_u_int32 ip;
};       

struct ban_string {
       char string[MAX_STRING];
       int port;
} banner_string;

 struct net_mask {
             unsigned long int start_dest;
             unsigned long int final_dest;
           } addr; 

struct icmp_net_mask {
	     unsigned long int start;	
	     unsigned long int finish;
} icmp_net;

struct host_list {
       char ip[BUF_LEN];
       int status;
       struct host_list *next;
};

struct os_list {
      char print[MAX_STRING];
      struct os_list *next;
};

FILE *osfd;
struct host_list *iplist;
struct os_list *oslist;

extern struct options opts;
extern struct local localhost;
extern pcap_t *p_handler;
extern int linktype; 

void fingerprint_(u_short, u_short, int, const u_char *);
int grep_string(u_long , int , char *);
int load_prints(char *);
void *icmp_scan(void *);
int icmp_packet(u_long);
void catch_sigint(int);
void add_ip(char *);
int get_mask(char *);
struct hostent * hostent_(char *);
int Logprint(char *fmt, ...);
void print_banner(void);
unsigned long int resolve_ip(char *, struct hostent *);
int ip_connect(unsigned long, int); 
char *recvbuf(int); 
int tcp_packet (u_long, u_long, u_short, u_short, u_short, u_short);
unsigned short in_cksum(unsigned short *, int);
int init_pcap_(char *, ...);
void pcap_done_(void);
int split_prints(char *, char *, char ***);
int get_tcp_ports(char *);
int port_connect(int, int);
int get_os(unsigned long, u_short);
int syn_portscan(unsigned long);
u_short sock_portscan(unsigned long);

/* Our threaded functions */ 

/* For single host we use a callback handler _checkp_ called by checkp() thread */

void *checkp(void *);
void _checkp_(u_char *pargs, const struct pcap_pkthdr *packetheader,
                          const u_char *packetdata);

/* Seperate callback handler for SYN portscan */

void *syn_handler(void *);
void _syn_handler(u_char *pargs, const struct pcap_pkthdr *packetheader,
			   const u_char *packetdata);

/* Seperate callback handler for icmp ping sweep */

void *icmp_handler(void *);
void _icmp_handler(u_char *pargs, const struct pcap_pkthdr *packetheader,
			   const u_char *packetdata);

/* For multiple hosts we use checkpackets with pcap_next() */
int recv_packets(unsigned long, char *);

  
void *scan(void *);
void _scan_(unsigned long );
 
typedef u_int32_t tcp_seq;

struct tcp_header {
        u_int16_t th_sport;             /* source port */
        u_int16_t th_dport;             /* destination port */
        tcp_seq   th_seq;               /* sequence number */
        tcp_seq   th_ack;               /* acknowledgement number */
#if BYTE_ORDER == LITTLE_ENDIAN
        u_int8_t  th_x2:4,              /* (unused) */
                  th_off:4;             /* data offset */
#endif
#if BYTE_ORDER == BIG_ENDIAN
        u_int8_t  th_off:4,             /* data offset */
                  th_x2:4;              /* (unused) */
#endif
        u_int8_t  th_flags;
#define TH_FIN    0x01
#define TH_SYN    0x02
#define TH_RST    0x04
#define TH_PUSH   0x08
#define TH_ACK    0x10
#define TH_URG    0x20
        u_int16_t th_win;                       /* window */
        u_int16_t th_sum;                       /* checksum */
        u_int16_t th_urp;                       /* urgent pointer */
};

struct ip_header {
#if BYTE_ORDER == LITTLE_ENDIAN
        u_int8_t  ip_hl:4,              /* header length */
                  ip_v:4;               /* version */
#endif
#if BYTE_ORDER == BIG_ENDIAN
        u_int8_t  ip_v:4,               /* version */
                  ip_hl:4;              /* header length */
#endif
        u_int8_t  ip_tos;               /* type of service */
        u_int16_t ip_len;               /* total length */
        u_int16_t ip_id;                /* identification */
        u_int16_t ip_off;               /* fragment offset field */
#define IP_RF 0x8000                    /* reserved fragment flag */
#define IP_DF 0x4000                    /* dont fragment flag */
#define IP_MF 0x2000                    /* more fragments flag */
#define IP_OFFMASK 0x1fff               /* mask for fragmenting bits */
        u_int8_t  ip_ttl;               /* time to live */
        u_int8_t  ip_p;                 /* protocol */
        u_int16_t ip_sum;               /* checksum */
        struct    in_addr ip_src, ip_dst; /* source and dest address */
};
struct icmp_header
{
  unsigned char icmp_type;
  unsigned char icmp_code;
  unsigned short int icmp_cksum;
  unsigned short int icmp_id;
  unsigned short int icmp_seq;
#define ICMP_ECHO 8
#define ICMP_ECHOREPLY 0
};
