
#include "abyss.h"

extern int local_ip;
/* # of icmp packets sent */
int nsent = 0;



unsigned long int
resolve_ip (char *host, struct hostent *hostname)
{
  unsigned long int ip_addr;

  if ((ip_addr = inet_addr (host)) != -1)
    return ip_addr;

  if ((hostname = gethostbyname (host)) == NULL)
    return -1;

  memcpy (&ip_addr, hostname->h_addr, hostname->h_length);
  return ip_addr;
}

struct hostent *
hostent_ (char *host)
{
  struct hostent *hostname;

  if ((hostname = gethostbyname (host)) == NULL)
    return -1;

  return (struct hostent *) hostname;
}

/* Create nonblocking socket */

int
ip_connect (unsigned long ip, int port)
{
  int sock, n;
  struct sockaddr_in abyss;
  u_long sflag = 0;

  if ((sock = socket (AF_INET, SOCK_STREAM, 0)) == -1)
    return -1;

  memset (&abyss, 0, sizeof (struct sockaddr_in));

  abyss.sin_family = AF_INET;
  abyss.sin_port = htons (port);
  abyss.sin_addr.s_addr = ip;

  sflag = fcntl (sock, F_GETFL, 0);
  fcntl (sock, F_SETFL, sflag | O_NONBLOCK);

  if ((n = connect (sock, (struct sockaddr *) &abyss, sizeof (abyss)) < 0))
    if (errno != EINPROGRESS)
      return -1;

  return sock;
}

int
port_connect (int sfd, int secs)
{
  int status, n;
  u_long len = 4;
  fd_set wset;
  struct timeval tval;

  FD_ZERO (&wset);
  FD_SET (sfd, &wset);

  tval.tv_sec = secs;
  tval.tv_usec = 0;

  select (sfd + 1, NULL, &wset, NULL, secs ? &tval : NULL);

  if (FD_ISSET (sfd, &wset))
    {
      if (getsockopt (sfd, SOL_SOCKET, SO_ERROR, &status, &len) < 0)
	return -1;
    }
  if (status == 0)
    return 1;
  else
    return 0;
}

char *
recvbuf (int sock)
{
  char *buf = (char *) malloc (strlen (MAX_STRING) + 1);
  char *p;
  int c;

  recv (sock, buf, MAX_STRING + 1, 0);

  if ((p = strchr (buf, 13)) != NULL)
    *(p) = '\0';

  return buf;
}

int
tcp_packet (u_long srcaddr, u_long dstaddr, u_short srcport, u_short dstport,
	    u_short flags, u_short win)
{

  char datagram[4096];
  char pseudohdr[1024];
  int tcpsize;
  int one = 1;
  int sockfd;
  const int *val = &one;

  struct pseudo
  {
    u_long srcaddr;
    u_long dstaddr;
    u_short length;
    u_char zero;
    u_char proto;
  };



  struct ip_header *iph = (struct ip_header *) datagram;
  struct pseudo *pseudo = (struct ip_header *) datagram;
  struct tcp_header *tcph =
    (struct tcp_header *) (datagram + sizeof (struct ip_header));

  struct sockaddr_in sin;
  struct sockaddr_in sout;

  tcpsize = sizeof (struct tcp_header);

  sin.sin_family = AF_INET;
  sin.sin_port = htons (dstport);
  sin.sin_addr.s_addr = dstaddr;

  memset (datagram, 0, 4096);

  iph->ip_hl = 5;
  iph->ip_v = 4;
  iph->ip_tos = 0;
  iph->ip_len = sizeof (struct ip_header) + sizeof (struct tcp_header);
  iph->ip_id = htons (getpid ());
  iph->ip_off = 0;
  iph->ip_ttl = 60;
  iph->ip_p = 6;
  iph->ip_sum = 0;
  iph->ip_src.s_addr = srcaddr;
  iph->ip_dst.s_addr = dstaddr;

  tcph->th_sport = srcport;
  tcph->th_dport = dstport;
  tcph->th_seq = htonl (rand ());
  tcph->th_ack = htonl (rand ());
  tcph->th_x2 = 0;
  tcph->th_off = 0;
  tcph->th_flags = flags;
  tcph->th_win = htons (win);
  tcph->th_sum =
    in_cksum ((u_short *) pseudo,
	      sizeof (struct pseudo) + sizeof (struct tcphdr));
  tcph->th_urp = 0;


  setsockopt (sockfd, IPPROTO_IP, IP_HDRINCL, val, sizeof (one));
  sendto (sockfd, datagram, iph->ip_len, 0, (struct sockaddr *) &sin,
	  sizeof (sin));

}
unsigned short
in_cksum (unsigned short *ptr, int nbytes)
{

  register long sum;		/* assumes long == 32 bits */
  u_short oddbyte;
  register u_short answer;	/* assumes u_short == 16 bits */

  /*
   * Our algorithm is simple, using a 32-bit accumulator (sum),
   * we add sequential 16-bit words to it, and at the end, fold back
   * all the carry bits from the top 16 bits into the lower 16 bits.
   */

  sum = 0;
  while (nbytes > 1)
    {
      sum += *ptr++;
      nbytes -= 2;
    }

  /* mop up an odd byte, if necessary */
  if (nbytes == 1)
    {
      oddbyte = 0;		/* make sure top half is zero */
      *((u_char *) & oddbyte) = *(u_char *) ptr;	/* one byte only */
      sum += oddbyte;
    }

  /*
   * Add back carry outs from top 16 bits to low 16 bits.
   */

  sum = (sum >> 16) + (sum & 0xffff);	/* add high-16 to low-16 */
  sum += (sum >> 16);		/* add carry */
  answer = ~sum;		/* ones-complement, then truncate to 16 bits */
  return (answer);
}

/* SENDS ICMP_ECHO */

void
ping_net (u_long ip)
{
  char *packet =
    malloc (sizeof (struct ip_header) + sizeof (struct icmp_header) + 50);
  struct sockaddr_in sockinfo;
  struct icmp_header *icmph;
  struct ip_header *iph;

  int i = 1, s;

  iph = (struct ip_header *) (packet);
  icmph = (struct icmp_header *) (packet + sizeof (struct ip_header));

  if ((s = socket (AF_INET, SOCK_RAW, IPPROTO_RAW)) == -1)
    {
      perror ("socket");
      exit (-1);
    }
  if (setsockopt (s, IPPROTO_IP, IP_HDRINCL, &i, sizeof (int)) == -1)
    {
      perror ("setsockopt");
      exit (-1);
    }
 
  iph->ip_src.s_addr = opts.myip;
  iph->ip_dst.s_addr = ip;
  iph->ip_ttl = 60;
  iph->ip_id = htons (getpid ());
  iph->ip_v = 4;
  iph->ip_hl = 5;
  iph->ip_tos = 0;
  iph->ip_p = IPPROTO_ICMP;

  icmph->icmp_type = ICMP_ECHO;
  icmph->icmp_code = 0;

  icmph->icmp_id = 666;
  icmph->icmp_seq = nsent++;
  icmph->icmp_cksum = 0;
  icmph->icmp_cksum = in_cksum ((u_short *) icmph,
				sizeof (struct icmp_header));

  iph->ip_len = sizeof (struct ip_header) + sizeof (struct icmp_header) + 50;

  sockinfo.sin_family = AF_INET;
  sockinfo.sin_addr.s_addr = iph->ip_dst.s_addr;

  if (sendto (s, packet, iph->ip_len,
	      0, (struct sockaddr *) &sockinfo,
	      sizeof (struct sockaddr)) == -1)
    {
      Logprint ("AbysS [Error] sendto() error\n");
      pthread_exit (pthread_self ());
    }


}

int
grep_string (u_long ip, int port, char *banner)
{
  int sockfd;
  u_char buf[4];
  int ret;
  char *string;

  if ((sockfd = ip_connect (ip, port)) < 0)
    return -1;

  if (opts.verbose)
    Logprint ("AbysS [VERBOSE] - grep_string-> Connected to ip %s\n",
	      inet_ntoa (ip));

  if ((ret = port_connect (sockfd, 1)) != 1)
    return -1;

  if (opts.verbose)
    Logprint ("AbysS [VERBOSE] - grep_string-> Port %d is open\n", port);

  if (port == 23)
    {
      if (read (sockfd, buf, 1) != 1)
	return -1;

      if (*buf == 255)
	{
	  read (sockfd, buf + 1, 2);

	  if (*(buf + 1) == 253)
	    {
	      *(buf + 1) = 252;
	      write (sockfd, buf, 3);
	    }
	}

      if ((*(buf + 1) < (u_char) 127) && (*(buf + 2) < (u_char) 127)
	  && (*(buf + 3) < (u_char) 127))
	{

	  string = recvbuf (sockfd);
	  if (strcmp (string, banner) == 0)
	    return 1;
	}
      else
	return -1;

    }
  else
    {
      string = recvbuf (sockfd);
      if (strcmp (string, banner) == 0)
	return 1;
      else
	return -1;
    }
}
