/* Handles scanning, and primary callback function */

#include "abyss.h"
extern unsigned long local_ip;
static int portscan = 0;

extern struct options opts;

void
_scan_ (unsigned long ip)
{
  char buf[4096];
  int i, err;
  unsigned long ipaddr;
  unsigned long local;

  struct current_host *ptr;

  struct packet
  {
    struct ip *ip;
    struct tcphdr *tcp;
    char data[2048];
  } *packet;

  if (opts.pingnet)
    {
      icmp_scan (ip);
      goto done;
    }

  if (opts.synscan)
    {
      ipaddr = ntohl (ip);
      syn_portscan (ipaddr);
      goto done;
    }

  if (opts.singlehost)
    {
      sock_portscan (ip);
      goto done;
    }

  else if ((opts.subscan) || (opts.fromlist))
    {
      /* add ip name to a linked list, so we can keep */
      /* track of what hosts we need to fingerprint */

      add_ip ((char *) inet_ntoa (ip));

      if ((err = sock_portscan (ip)) == -1)
	goto done;
    }

done:
  return;

}

/* For -s singlehost option, we have a thread call checkp() as a dummy function */
/* Which then uses pcap_loop() on _checkp_ as a call back function waiting for packets */

void *
checkp (void *arg)
{
  /* We call our _checkp_ function as a packet handler by pcap_loop */
  /* second arg -1 means read packets forever or until an error occurs */

  if ((pcap_loop (p_handler, -1, _checkp_, NULL)) < 0)
    {
      Logprint ("AbysS [PCAP] pcap_loop error %s\n", errbuf);
      exit (-1);
    }
  return NULL;
}

/* This is our main callback handler */
/* Keeps track of what ip's have already had OS checks */
/* The ones that havent, get fingerprinted */

void
_checkp_ (u_char * pargs, const struct pcap_pkthdr *packetheader,
	  const u_char * packetdata)
{

  struct host_list *current = iplist;

  /* Will be checking these for prints */
  u_short win, ttl;
  int df;
  unsigned int off;

  char *data, *tmp;
  char buf[500], flags[50], *flarray[10];
  char *ptr;
  char *databuf;
  char *datastr;
  int j, c, z, flag_len, k = 0;
  int data_len;
  int count, cnt;
  int dummy;
  int i = 0;

  u_short dport, sport;

  struct ether_header *ether;
  struct ip_header *ip;
  struct tcp_header *tcp;

  ether = (struct ether_header *) packetdata;
  ip = (struct ip *) (packetdata + sizeof (struct ether_header));
  tcp =
    (struct tcphdr *) (packetdata + sizeof (struct ether_header) +
		       (ip->ip_hl * 4));

  data = (char *) ((int) tcp + (int) (tcp->th_off * 4));

  ttl = ip->ip_ttl;
  win = ntohs (tcp->th_win);
  off = ntohs (ip->ip_off);

  sport = ntohs (tcp->th_sport);
  dport = ntohs (tcp->th_dport);

  if (off & IP_DF)
    df = 1;
  else
    df = 0;

  if (ip->ip_p == IPPROTO_TCP)
    {
      if (opts.passivemode)
	{
	  if (opts.headers)
	    {

	      data_len = strlen (data);

	      if ((databuf = (char *) malloc (data_len + 1)) == NULL)
		{
		  printf ("Could not initialize memory!\n");
		  pthread_exit (pthread_self ());
		}
	      if ((datastr = (char *) malloc (data_len + 1)) == NULL)
		{
		  printf ("Could not initialize memory!\n");
		  pthread_exit (pthread_self ());
		}
	      memcpy (databuf, data, data_len);
	      cnt = 0;


	      for (count = 0; count < data_len; count++)
		{
		  if ((isalnum (databuf[count]) || ispunct (databuf[count])
		       || databuf[count] == ' ') && (count < data_len))
		    {
		      datastr[cnt] = databuf[count];
		      datastr[cnt + 1] = '\0';
		      cnt++;

		    }
		  else if ((isspace (databuf[count]) && databuf[count] != ' ')
			   && (cnt < data_len - 1))
		    {
		      datastr[cnt] = ' ';
		      datastr[cnt + 1] = 0;
		      cnt++;
		    }
		}

	      if ((tmp = (char *) malloc (50)) == NULL)
		{
		  printf ("Could not allocate memory!\n");
		  exit (-1);
		}
	      k = 0;
	      for (j = 0; j < i; j++)
		{
		  c = strlen (flarray[j]);
		  strcpy (tmp, flarray[j]);
		  for (z = 0; z < c; z++)
		    flags[k++] = tmp[z];
		  flags[k++] = '/';
		  memset (tmp, 0, sizeof (tmp));
		}

	      flags[k] = '\0';

	      Logprint ("\nsrc  [%s] -> ", inet_ntoa (ip->ip_src.s_addr));
	      Logprint ("dst  [%s] -> ", inet_ntoa (ip->ip_dst.s_addr));
	      Logprint
		("sport %d -> dport %d -> \nttl %d -> win %d -> off %d -> df %d -> flags: %c%c%c%c%c%c\n\n",
		 sport, dport, ttl, win, off, df,
		 (tcp->th_flags & TH_URG) ? 'U' : '-',
		 (tcp->th_flags & TH_ACK) ? 'A' : '-',
		 (tcp->th_flags & TH_PUSH) ? 'P' : '-',
		 (tcp->th_flags & TH_RST) ? 'R' : '-',
		 (tcp->th_flags & TH_SYN) ? 'S' : '-',
		 (tcp->th_flags & TH_FIN) ? 'F' : '-');

	      Logprint ("%s", datastr);

	      for (j = 0; j < i; j++)
		free (flarray[j]);

	      free (databuf);
	      free (datastr);
	      i = 0;
	    }
	}

      if (tcp->th_flags == (TH_SYN | TH_ACK))
	{
          if (opts.passivemode)
          Logprint("AbysS [Report] - Received SYN/ACK ... Analyzing fingerprints\n");
	  if (opts.singlehost || opts.passivemode)
	    fingerprint_ (win, ttl, df, packetdata); 
	  sleep (1);
	}
    }

  if (ip->ip_p == IPPROTO_TCP)
    {
      if (tcp->th_flags == (TH_SYN | TH_ACK))
	{

	  if ((opts.subscan) || (opts.fromlist))
	    {

	      while (current != 0)
		{
		  if (strcmp (current->ip, inet_ntoa (ip->ip_src.s_addr)) ==
		      0)
		    {
		      if (current->status == 0)
			{
			  fingerprint_ (win, ttl, df, packetdata);
			  current->status = 1;
			  break;
			}
		    }
		  current = current->next;
		}
	    }
	}
    }

  usleep (350000);

}
