#include "abyss.h"

pthread_mutex_t write_lock = PTHREAD_MUTEX_INITIALIZER;

/* NOTE, the mutex locks around calls to Logprint() are their */
/* to do the same thing as flockfile() only because the Logprint fd */
/* is not global, otherwise we *would* use flockfile */

void
fingerprint_ (u_short win, u_short ttl, int df, const u_char * packetdata)
{

  struct host_list *current = iplist;

  /* parse the fingerprint into tokens with our split_prints() */
  char **fingerprints, *string, buf[BUF_LEN], *ptr;

  /* os, ttl, win, df */
  char *pos, *ptl, *pwn, *pdf;

  char final_os[BUF_LEN];

  FILE *fd;
  unsigned int off;
  int tokens, i = 0;
  int real_match = 0;
  int j = 0;
  int ttl_diff;
  char *c;

  struct saved_os
  {
    char os[BUF_LEN];
    int ttl;
  } saved_os[10];


  struct guess_os
  {
    char os[BUF_LEN];
    int ttl;
  } guess_os[10];

  int saved_os_num = 0;
  int guess_os_num = 0;

  int ttl_subtract[10];

  struct ether_header *ether;
  struct ip_header *ip;
  struct tcp_header *tcp;

  ether = (struct ether_header *) packetdata;
  ip = (struct ip *) (packetdata + sizeof (struct ether_header));
  tcp =
    (struct tcphdr *) (packetdata + sizeof (struct ether_header) +
		       (ip->ip_hl * 4));



  if ((fd = fopen ("os.prints", "r")) == NULL)
    {
      Logprint ("AbysS [Error] - Could not read fingerprint file\n");
      pthread_exit (pthread_self ());
    }

  if (ip->ip_p == IPPROTO_TCP)
    {
      if (tcp->th_flags == (TH_SYN | TH_ACK))
	{

	  if (opts.verbose)
	    {
	      pthread_mutex_lock (&write_lock);
	      Logprint
		("AbysS [Report] - NEW PACKET - ANALYZING FINGERPRINTS - \n");
	      Logprint ("WindowSize [%d]\n", win);
	      Logprint ("TimeToLive [%d]\n", ttl);
	      Logprint ("DF(Ragment [%d]\n", df);
	      Logprint ("AbysS [Report] - Done!\n");
	      pthread_mutex_unlock (&write_lock);

	    }

	  flockfile (fd);
	  while ((fgets (buf, BUF_LEN, fd)) != NULL)
	    {
	      funlockfile (fd);

	      /* split_prints (parse.c) splits the fingerprints into an argument vector */

	      if ((tokens = split_prints (buf, ":", &fingerprints)) > 0)
		{
		  pwn = fingerprints[0];
		  pdf = fingerprints[1];
		  ptl = fingerprints[2];
		  pos = fingerprints[3];

		  if (win == atoi (pwn) && df == atoi (pdf))
		    {
		      if (ttl > atoi (ptl))
			{

			  if (opts.verbose)
			    {

			      pthread_mutex_lock (&write_lock);
			      Logprint
				("AbysS [OS-Detect] - Warning. TTL Exceeds atleast one match\n");
			      pthread_mutex_unlock (&write_lock);
			    }
			  strcpy (guess_os[j].os, pos);
			  guess_os[i].ttl = atoi (ptl);
			  j++;
			  guess_os_num++;	/* use guess_os_num for ttl's larger */
			  continue;	/* than real ttl                     */
			}
		      else if (ttl < atoi (ptl))
			{
			  strcpy (saved_os[i].os, pos);
			  saved_os[i].ttl = atoi (ptl);
			  i++;
			  saved_os_num++;	/* use saved_os_num for ttl's smaller */
			  continue;	/* than real ttl                      */
			}
		      else if (ttl == atoi (ptl))
			{
			  strcpy (final_os, pos);
			  real_match = 1;
			  if (opts.verbose)
			    {
			      pthread_mutex_lock (&write_lock);
			      Logprint
				("AbysS - [OS-Detect] - Found perfect OS match!\n");
			      pthread_mutex_unlock (&write_lock);
			    }
			  goto final;
			}
		    }
		}

	    }

	  /* We must guess the most accurate ttl value   */
	  /* Abyss does not use a U1 probe, as does nmap  */
	  /*  to find out how many hops away the target is */

	  /* this first part is the inauspicious guess */
	  if (guess_os_num > 0 && saved_os_num < 1 && real_match != 1)
	    {

	      if (guess_os_num > 1)
		{
		  for (i = 0; i < guess_os_num; i++)
		    {
		      if (ttl >= guess_os[i].ttl)
			ttl_subtract[i] = ttl - guess_os[i].ttl;
		    }

		  for (j = 0; j < guess_os_num - 1; j++)
		    {
		      if (ttl_subtract[j] < ttl_subtract[j + 1])
			ttl_diff = ttl_subtract[j];
		      else
			ttl_diff = ttl_subtract[j + 1];
		      j++;
		    }
		  for (i = 0; i < guess_os_num; i++)
		    {
		      if ((ttl - ttl_diff) == guess_os[i].ttl)
			strcpy (final_os, guess_os[i].os);
		    }
		}
	      else
		strcpy (final_os, guess_os[0].os);
	    }

	  /* this part is the superior guess and is probably right */
	  if (saved_os_num > 0 && real_match != 1)
	    {

	      if (saved_os_num > 1)
		{
		  for (i = 0; i < saved_os_num; i++)
		    {
		      if (ttl < saved_os[i].ttl)
			ttl_subtract[i] = saved_os[i].ttl - ttl;
		    }

		  for (j = 0; j < saved_os_num - 1; j++)
		    {
		      if (ttl_subtract[j] < ttl_subtract[j + 1])
			ttl_diff = ttl_subtract[j];
		      else
			ttl_diff = ttl_subtract[j + 1];
		    }

		  for (i = 0; i < saved_os_num; i++)
		    {

		      if ((ttl_diff + ttl) == saved_os[i].ttl)
			strcpy (final_os, saved_os[i].os);

		    }
		}
	      else
		strcpy (final_os, saved_os[0].os);

	    }

	  else
	    /* win, df, ttl do not match at all */

	  if (saved_os_num < 1 && guess_os_num < 1 && real_match != 1)
	    strcpy (final_os, "Unknown OS");

	final:

	  if ((c = strchr (final_os, '\n')) != NULL)
	    *(c) = '\0';

	  /* Logprint() uses a file descriptor, but it is not global */
	  /* so we dont use flockfile() instead we use a mutex */

	  //pthread_mutex_lock(&write_lock);

	  if (opts.passivemode)
	    {
	      pthread_mutex_lock (&write_lock);
	      Logprint ("AbysS [Report] -> (OS Fingerprint Match)\n");
	      Logprint ("AbysS [Report] -> [ %s ] -> [%s]\n",
			inet_ntoa (ip->ip_src.s_addr), final_os);
	      pthread_mutex_unlock (&write_lock);
	      fclose (fd);
	    }

	  if (opts.singlehost)
	    {
	      pthread_mutex_lock (&write_lock);
	      Logprint ("AbysS [Report] -> (OS Fingerprint Match )\n");
	      Logprint ("AbysS [Report] -> [ %s ] -> OS [%s]\n",
			inet_ntoa (ip->ip_src.s_addr), final_os);
	      pthread_mutex_unlock (&write_lock);
	      fclose (fd);
	      pthread_exit (pthread_self ());
	    }

	  if (opts.subscan || opts.fromlist)
	    {
	      pthread_mutex_lock (&write_lock);
	      Logprint ("AbysS [Report] -> ( OS Fingerprint Match )\n");
	      Logprint ("AbysS [Report] -> [ %s ] -> [OS] [%s]\n",
			inet_ntoa (ip->ip_src.s_addr), final_os);
	      pthread_mutex_unlock (&write_lock);
	      fclose (fd);
	    }


	}
    }
}
